//! Acme client that supports multiple CAs and configs for sites that can be seperate from the mainconfig
#![allow(dead_code)]
#![allow(clippy::clone_on_copy)]
#![allow(clippy::identity_op)]

pub(crate) mod consts;
pub(crate) mod macros;
pub(crate) mod prelude;
pub(crate) mod process;
pub(crate) mod types;
pub(crate) mod utils;

use crate::{
    consts::*,
    prelude::*,
    process::{
        process_accounts,
        process_site,
        services,
    },
    types::{
        config::{
            GeneralConfig,
            SiteConfig,
        },
        structs::{
            Arguments,
            ProcessorArgs,
        },
    },
};
use acme2_eab::Directory;
use async_scoped::TokioScope;
use clap::Parser;
use env_logger::init as log_init;
use libsystemd::daemon;
use log::*;
use openssl::{
    self,
    pkey::{
        PKey,
        Private,
    },
};
use reqwest::tls::Version;
use std::{
    collections::{
        HashMap,
        HashSet,
    },
    path::{
        Path,
        PathBuf,
    },
    str::FromStr,
    sync::Arc,
    time::Duration,
};
use tokio::{
    fs::{
        create_dir_all,
        read_dir,
    },
    io::AsyncReadExt,
    sync::Mutex,
};
use tokio_stream::{
    StreamExt,
    wrappers::ReadDirStream,
};


fn default_client() -> reqwest::Client {
    reqwest::Client::builder().min_tls_version(Version::TLS_1_2).https_only(true).pool_max_idle_per_host(POOL_SIZE).build().unwrap()
}

async fn load_privkey(path: PathBuf) -> Result<PKey<Private>, ()> {
    let mut file = match_error!(FILE_MODE.open(path).await=>Err(error)-> "Failed to open Private Key: {error}", Err(()));
    let mut data = String::new();
    if let Err(error) = file.read_to_string(&mut data).await {
        error!("Failed to read data for the key: {error}");
        return Err(());
    }
    match PKey::private_key_from_pem(data.as_bytes()) {
        Ok(key) => Ok(key),
        Err(error) => {
            error!("Failed to parse pem data: {error}");
            Err(())
        },
    }
}

async fn racme(flags: Arguments) {
    let client = default_client();
    let systemd_access = daemon::booted();
    let mainconfig = GeneralConfig::from_file(match_error!(FILE_MODE.open(flags.config).await=>Err(error)-> "error reading the config: {error}")).await;
    trace!("Parsed Config: {mainconfig:?}");
    let files = ReadDirStream::new(match_error!(read_dir(mainconfig.sites_path.clone()).await=>Err(error)-> "could not read files from sites dir: {error}"));
    let mut siteconfigs = Vec::new();
    for file in files.filter(Result::is_ok).map(|file| file.unwrap().path()).collect::<Vec<PathBuf>>().await {
        let mut site = SiteConfig::from_file(FILE_MODE.open(file.clone()).await.unwrap()).await;
        site.name = file.file_stem().unwrap().to_str().unwrap().to_string();
        siteconfigs.push(site);
    }
    let used = siteconfigs.iter().map(|s| s.ca.clone()).collect::<HashSet<_>>();
    debug!("Used CAs: {used:?}");

    let mut directories = HashMap::<String, Arc<Directory>>::new();

    let mut accounts = HashMap::new();
    let accountpath = Path::new(&mainconfig.accounts_path).to_path_buf();
    if let Err(error) = create_dir_all(accountpath.clone()).await {
        error!("Failed to create the directory for the accounts: {error}");
        return;
    }
    for (name, ca) in mainconfig.ca.iter().filter(|(name, _)| used.contains(name.to_owned())) {
        process_accounts(name, ca, &mut directories, &mut accounts, &client, accountpath.clone()).await;
    }


    let restart_services = Mutex::new(HashSet::<String>::new());
    let reload_services = Mutex::new(HashSet::<String>::new());
    let certs = Path::new(&mainconfig.certificates_path).to_path_buf();

    if !certs.exists() {
        if let Err(error) = create_dir_all(certs.clone()).await {
            error!("Failed to create directory for all the certificates: {error}");
            return;
        }
    }
    let challengepath = mainconfig.http_challenge_path.and_then(|path| PathBuf::from_str(path.as_str()).ok());
    let dnsserver = None;

    unsafe {
        TokioScope::scope_and_collect(|scope| {
            for site in siteconfigs {
                if let Some(account) = accounts.get(&site.ca) {
                    scope.spawn(process_site(ProcessorArgs::new(
                        site,
                        Arc::clone(account),
                        &reload_services,
                        &restart_services,
                        certs.clone(),
                        challengepath.clone(),
                        dnsserver.clone(),
                    )));
                } else {
                    error!("Could not process site {} because of previous errors", site.name)
                }
            }
        })
    }
    .await;

    if systemd_access {
        services(restart_services.into_inner(), reload_services.into_inner()).await;
    }
}

fn main() {
    log_init();
    let runtime = tokio::runtime::Builder::new_current_thread().enable_all().build().unwrap();
    runtime.block_on(racme(Arguments::parse()));
    runtime.shutdown_timeout(Duration::from_secs(1));
}
