use std::collections::HashSet;

use crate::consts::{
    BRAINPOOL_MIDDLE,
    BRAINPOOL_STRONG,
    BRAINPOOL_WEAK,
    SECP_MIDDLE,
    SECP_STRONG,
    SECP_WEAK,
};
use log::*;
use openssl::{
    asn1::Asn1Time,
    pkey::{
        Id,
        PKey,
        Private,
    },
    x509::X509,
};
use serde::de::DeserializeOwned;
use tokio::{
    fs::File,
    io::AsyncReadExt,
};

use super::{
    cryptography::{
        Algorithm,
        Strength,
    },
    structs::San,
};

pub trait FromFile: Default + DeserializeOwned {
    async fn from_file(file: File) -> Self;
}

impl<T: Default + DeserializeOwned> FromFile for T {
    async fn from_file(mut file: File) -> Self {
        let mut data = String::new();

        match file.read_to_string(&mut data).await {
            Ok(_) => {},
            Err(error) => {
                warn!("Failed to load config: {error}");

                return Default::default();
            },
        }

        match toml::from_str(&data) {
            Ok(output) => output,
            Err(error) => {
                warn!("Failed to parse toml file: {error}");

                Default::default()
            },
        }
    }
}


pub trait MatchAlgorithm {
    fn matches(&self, algorithm: Algorithm, strength: Strength) -> bool;
}

impl MatchAlgorithm for PKey<Private> {
    fn matches(&self, algorithm: Algorithm, strength: Strength) -> bool {
        match (self.id(), algorithm) {
            (Id::ED25519, Algorithm::ED25519) => true,
            (Id::RSA, Algorithm::Rsa) if self.bits() == strength.rsabits() => true,
            (Id::EC, Algorithm::Secp) | (Id::EC, Algorithm::Brainpool) => {
                let pkey = self.ec_key().unwrap();
                let curve = pkey.group().curve_name().unwrap();
                match (algorithm, strength) {
                    (Algorithm::Secp, Strength::Weak) if SECP_WEAK == curve => true,
                    (Algorithm::Secp, Strength::Middle) if SECP_MIDDLE == curve => true,
                    (Algorithm::Secp, Strength::Strong) if SECP_STRONG == curve => true,
                    (Algorithm::Brainpool, Strength::Weak) if BRAINPOOL_WEAK == curve => true,
                    (Algorithm::Brainpool, Strength::Middle) if BRAINPOOL_MIDDLE == curve => true,
                    (Algorithm::Brainpool, Strength::Strong) if BRAINPOOL_STRONG == curve => true,
                    _ => false,
                }
            },
            _ => false,
        }
    }
}

pub trait MatchX509 {
    fn days_left(&self, days: u32) -> bool;
    fn match_san<T: IntoIterator<Item = San>>(&self, names: T) -> bool;
}

impl MatchX509 for X509 {
    fn days_left(&self, days: u32) -> bool {
        let future_date = Asn1Time::days_from_now(days).unwrap();
        self.not_after().compare(&future_date).is_ok_and(|order| order.is_le())
    }

    fn match_san<T: IntoIterator<Item = San>>(&self, names: T) -> bool {
        let alt_names = match self.subject_alt_names() {
            None => return false,
            Some(x) => x,
        };
        let mut cert_san = HashSet::<San>::new();
        for san in alt_names {
            cert_san.insert(san.into());
        }
        let mut config_san = HashSet::new();
        for san in names {
            config_san.insert(san);
        }
        config_san.difference(&cert_san).count() == 0
    }
}
